<?php

/**
 * @file similarity.inc
 * lovingly crafted by Scott Reynolds
 *
 * Provides the generic factory methods for loading a similarity object and outlines
 * out the abstract classes for node similarities and similarities in general
 */

/**
 * Abstract klass defining the basics of the similarity class
 */
abstract class similarity {
  // internal id used as a primary key
  public $sid = 0;
  
  // the type of similarity. shall be implemented in overriding classes
  public $type;
  
  // Human readable title of the similarity class
  public $title;
  
  // machine readable name
  public $machine_name;
  
  // serialized data
  public $data;
  
  // calculate the similarities for the type
  // TODO: use the calculation_limits function to get the limits
  abstract public function calculate();
  
  // generate the second step of the form
  // defining whatever is needed by the class
  public function options_form() {
    return array(
      'title' => array(
        '#type' => 'textfield',
        '#title' => t('Name'),
        '#description' => t('Human readable title of the similary object. This name will be visible only on Adminstration pages'),
        '#required' => TRUE,
        '#default_value' => $this->title,
        '#size' => 40,
      ),
      'min_sim' => array(
        '#type' => 'select',
        '#title' => t('Minimum similarity score'),
        '#options' => drupal_map_assoc(range(50, 100, 5)),
        '#description' => t('Only store similarities greater then this number. This means that if a node is similar to another node to the 50%, it will only be stored if this value is 50 or greater'),
        '#default_value' => isset($this->min_sim) ? $this->min_sim : 50,
      ),
      'time_limit' => array(
        '#type' => 'select',
        '#title' => t('Time Limit'),
        '#description' => t('The number of minutes allowed for processing of similarities.'),
        '#options' => drupal_map_assoc(array(60, 60 * 2, 60 * 3, 60 * 4, 60 * 5, 60 * 6, 60 * 7, 60 * 8, 60 * 9, 60 * 10), 'format_interval')
      ),
    );
  }
  
  // table schema
  abstract protected function schema();
  
  // views data
  abstract public function views_data();
  
  /**
   * set the limits for calculation
   *
   * @return
   * array of limits for processing
   */
  public function calculation_limits() {
    // TODO: implement this. See notifications/messaging framework
    return array('number' => 100);
  }
  
  /**
   * basic constructor
   * This constructor maintains the similarity_objects table
   *
   * @param sid
   * sid can be null allows for the creation of a blank similarity object
   * and then add in the fields (minus sid) by hand, and a save()
   */
  public function __construct($sid = NULL) {
    if (!empty($sid)) {
      $db_object = db_fetch_object(db_query("SELECT * FROM {similarity_objects} WHERE sid = %d", $sid));
      foreach($db_object as $field => $value) {
        $this->$field = $value;
      }
      
      // grab the stuff and unserialize the data field
      drupal_unpack($this, 'data');
    }
  }
  
  /**
   * save the similarity object to the database
   */
  public function save() {
    // pass the proper update param to drupal_write_record
    $update = ($this->sid) ? 'sid' : NULL;
    
    $success = drupal_write_record('similarity_objects', $this, $update);
    if (!$success) {
      drupal_set_message(t('Failed to save the similarity object to the database'), 'error');
    }
    
    if ($success == SAVED_NEW) {
      // for semantics, we now have an sid
      $this->sid = db_last_insert_id('similarity_objects', 'sid');
      
      // create the objects table
      $ret = array();
      $schema = $this->schema();
      db_create_table($ret, $this->machine_name . '_similarity', $schema);
    }
  }
    
  /**
   * save the serialized verision of the second step of creation
   */
  public function serialize($form_values) {    
    // remove the things we know we don't want saved
    unset($form_values['similarity_object']);
    unset($form_values['op']);
    unset($form_values['submit']);
    unset($form_values['form_build_id']);
    unset($form_values['form_token']);
    unset($form_values['form_id']);
    
    $this->data = serialize($form_values);
  }
}

/**
 * Class similarity_node is used as an abstract class for all node
 * similarities
 */
abstract class similarity_node extends similarity {  
  /**
   * provide a table schema
   */
  protected function schema() {
    $schema = array(
      'fields' => array(
        'nid1' => array(
          'type' => 'int',
          'size' => 'normal',
          'not null' => TRUE,
        ),
        'nid2' => array(
          'type' => 'int',
          'size' => 'normal',
          'not null' => TRUE,
        ),
        'score' => array(
          'type' => 'float',
          'not null' => TRUE,
        ),
        'created' => array(
          'descritpion' => t('Used as a timestamp from when the entry was last created'),
          'type' => 'int',
          'not null' => TRUE,
        ),
      ),
      'primary key' => array('nid1', 'nid2'),
      'indexes' => array(
        'id_1' => array('nid1'),
        'id_2' => array('nid2'),
        'score' => array('score'),
      ),
    );
    
    return $schema;
  }
  
  /**
   * Implementation of the option form function
   *
   * @return
   * structured array (Drupal form)
   */
  public function options_form() {
    $form = parent::options_form();
    $node_types = node_get_types('names');
    $form['node_limit'] = array(
      '#type' => 'select',
      '#title' => t('Node Throttle'),
      '#options' => drupal_map_assoc(range(50, 400, 25)),
      '#description' => t('Select the number of node similarities to calculate per cron run. The higher the throttle the longer your cron process will take. If you notice your cron process taking too long, lower this number and consider changing the node types'),
      '#default_value' => isset($this->node_limit) ? $this->node_limit : 100,
    );
    $form['types'] = array(
      '#type' => 'checkboxes',
      '#title' => t('Node types'),
      '#description' => t('Select the node types that should be used to calculate similarities'),
      '#options' => $node_types,
      '#default_value' => isset($this->types) ? $this->types : array(),
    );
    return $form;
  }
  
  /**
   * override of serialize function
   */
  public function serialize(&$form_values) {
    $types = array();
    // silly checkboxes
    foreach (array_values($form_values['types']) as $type) {
      if (!empty($type)) {
        $types[] = $type;
      }
    }
    $form_values['types'] = $types;
    parent::serialize($form_values);
  }
  
  /**
   * returns the structure array for its tables
   */
  function views_data() {
    $table_name = $this->machine_name . '_similarity';
    
    $data[$table_name]['table']['group'] = t('Similarity');
    $data[$table_name]['table']['join'] = array(
      'node' => array(
        'left_field' => 'nid',
        'field' => 'nid1',
        'type' => 'LEFT',
      ),
    );
     $data[$table_name . "_reverse"]['table']['group'] = t('Similarity');
    $data[$table_name . "_reverse"]['table']['join'] = array(
      'node' => array(
        'left_field' => 'nid',
        'field' => 'nid2',
        'type' => 'LEFT',
      ),
    );
    $data[$table_name]['nid1'] = array(
      'title' => t('@name Similary NID', array('@name' => $this->title)),
      'argument' => array(
        'handler' => 'similarity_views_argument_nid',
      ),
      'help' => t('Node id of the similar node'),
    );
    
    $data[$table_name]['score'] = array(
      'title' => t('@name Similarity score', array('@name' => $this->title)),
      'field' => array(
        'handler' => 'similarity_views_handler_field_score', // TODO: write this handler
        'click sortable' => TRUE,
      ),
      'sort' => array(
        'handler' => 'views_handler_sort',
      ),
      'help' => t('Similarity score of the second node'),
    );

    return $data;
  }
  
  /**
   * return a type string for the SQL
   */
  protected function node_type_string() {
    $type_string = '';
    
    // yuck, can't enforce checkboxes as required. Thus...
    if (empty($this->types)) {
      return $type_string;
    }
    
    foreach($this->types as $node_type) {
      if (empty($type_string)) {
        $type_string = "n.type = '%s'";
      }
      else {
        $type_string .= " OR n.type = '%s'";
      }
    }
    $type_string = "(" . $type_string . ")";
    return $type_string;
  }
}

// util functions
/**
 * calculates teh cosine between objects a and b
 *
 * @param $a
 * array of revelancy scores keyed by word
 * ie 'wizard' => 7.33333
 *
 * @param $b
 * array of revelancy scores keyed by word
 * ie 'wizard' => 7.33333
 * 
 * @see get_words
 */
function similarity_cosine($a, $b) {
  $dot_product = 0;
  $a_total = 0;
  $b_total = 0;
  
  $total_words = array_unique(array_merge(array_keys($a), array_keys($b)));

  // only care about the key $word
  foreach($total_words as $word) {
    if (isset($a[$word]) && isset($b[$word])) {
      $dot_product += $a[$word] * $b[$word];
    }
    
    if (isset($a[$word])) {
      $a_total += pow($a[$word], 2);
    }
    
    if (isset($b[$word])) {
      $b_total += pow($b[$word], 2);
    }
  }
  
  // if nothing worked
  if ($a_total == 0 || $b_total == 0 ) {
    return 0;
  }
  
  return ($dot_product / sqrt($a_total * $b_total));
}
